# most of the drivers only need a couple of these... moved all up here for clarity below
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import (
        Unpack,  # can be imported from typing if python >= 3.12
    )

from qcodes import validators as vals
from qcodes.instrument import (
    VisaInstrument,
    VisaInstrumentKWArgs,
)
from qcodes.parameters import Parameter
from qcodes.validators import Enum, Numbers




class Keithley2182A1ch(VisaInstrument):
    """
    Instrument Driver for Keithley2182A (1 channel, Voltage only)

    Attributes:
        nplc (Parameter): Set or get the number of power line cycles (min=0.01, max=50)
        auto_range (Parameter): Set or get the measurement range automatically (1: ON, 0: OFF)
        rel (Parameter): Enables or disables the application of
                         a relative offset value to the measurement. (1: ON, 0: OFF)
        active (Parameter): Set or get the active function. (VOLT or TEMP)
        filter (Parameter): Enables or disables the digital filter for measurements.
        amplitude (Parameter): Get the voltage (unit: V)

    """
    def __init__(
        self,
        name: str,
        address: str,
        reset: bool = False,
        **kwargs: "Unpack[VisaInstrumentKWArgs]",
    ):

        super().__init__(name, address, **kwargs)

        self._trigger_sent = False

        self.nplc: Parameter = self.add_parameter(
            "nplc",
            get_cmd="SENS:VOLT:NPLC?",
            set_cmd="SENS:VOLT:NPLC {}",
            vals=Numbers(min_value=0.01, max_value=50),
            get_parser=float
        )

    
        self.activate: Parameter = self.add_parameter(
            "active",
            get_cmd=":SENS:FUNC?",
            set_cmd="SENS:FUNC {}",
            vals=Enum("VOLT", "TEMP")
        )

        self.channnel: Parameter = self.add_parameter(
            "channel",
            get_cmd=":SENS:CHAN?",
            set_cmd="SENS:CHAN {}",
            vals=Enum(1, 2)
        )

        self.filter: Parameter = self.add_parameter(
            "filter",
            get_cmd=":SENS:VOLT:DFILter:STAT?",
            set_cmd=":SENS:VOLT:DFILter:STAT {}"
        )

        self.amplitude: Parameter = self.add_parameter(
            "amplitude",
            get_cmd="SENS:DATA:FRES?",
            get_parser=float,
            unit="V"
        )

        self.get = self.amplitude


class Keithley2182AChannel(InstrumentChannel):
    """
    Class to hold the two Keithley channels, i.e.
    CH1 and CH2.
    """

    def __init__(
        self,
        parent: Instrument,
        name: str,
        channel: str,
        **kwargs: "Unpack[InstrumentBaseKWArgs]",
    ) -> None:
        """
        Args:
            parent: The Instrument instance to which the channel is
                to be attached.
            name: The 'colloquial' name of the channel
            channel: The name used by the Keithley, i.e. either
                'ch1' or 'ch2'
            **kwargs: Forwarded to base class.

        """

        if channel not in ["ch1", "ch2"]:
            raise ValueError('channel must be either "ch1" or "ch2"')

        dchannel = {
            "ch1": "CHAN1",
            "ch2": "CHAN2"
        }

        super().__init__(parent, name, **kwargs)
        self.model = self._parent.model

        self.auto_range: Parameter = self.add_parameter(
            "auto_range",
            get_cmd=f"SENS:VOLT:{dchannel[channel]}:RANG:AUTO?",
            set_cmd=f"SENS:VOLT:{dchannel[channel]}:RANG:AUTO {{}}"
        )

        self.vrel: Parameter = self.add_parameter(
            "vrel",
            get_cmd=f"SENS:VOLT:{dchannel[channel]}:REFerence:STATe?",
            set_cmd=f"SENS:VOLT:{dchannel[channel]}:REFerence:STATe {{}}"
        )

        self.amplitude: Parameter = self.add_parameter(
            "amplitude",
            get_cmd="SENS:DATA:FRES?",
            get_parser=float,
            unit="V"
        )

        



class Keithley2600(VisaInstrument):
    """
    This is the qcodes driver for the Keithley2600 Source-Meter series,
    tested with Keithley2614B
    """

    default_terminator = "\n"

    def __init__(
        self, name: str, address: str, **kwargs: "Unpack[VisaInstrumentKWArgs]"
    ) -> None:
        """
        Args:
            name: Name to use internally in QCoDeS
            address: VISA ressource address
            **kwargs: kwargs are forwarded to the base class.

        """
        super().__init__(name, address, **kwargs)

        model = self.ask("localnode.model")

        knownmodels = [
            "2601B",
            "2602B",
            "2604B",
            "2611B",
            "2612B",
            "2614B",
            "2635B",
            "2636B",
        ]
        if model not in knownmodels:
            kmstring = ("{}, " * (len(knownmodels) - 1)).format(*knownmodels[:-1])
            kmstring += f"and {knownmodels[-1]}."
            raise ValueError("Unknown model. Known model are: " + kmstring)

        # Add the channel to the instrument
        self.smua = self.add_submodule("smua", KeithleyChannel(self, "smua", "smua"))
        "smua channel of the instrument"
        self.smub = self.add_submodule("smub", KeithleyChannel(self, "smub", "smub"))
        "smub channel of the instrument"

        # display parameter
        # Parameters NOT specific to a channel still belong on the Instrument object
        # In this case, the Parameter controls the text on the display
        self.display_settext = Parameter(
            "display_settext",
            set_cmd=self._display_settext,
            vals=vals.Strings(),
            instrument=self,
        )

        self.connect_message()


from collections.abc import Collection
from numbers import Real

import pandas as pd


def critical_current(current: Collection[Real], resistance: Collection[Real], threshold: Real | None = None) -> float:
    """
    Calculate the critical current from a current-resistance curve.

    Args:
        current: The current values.
        resistance: The resistance values.
        threshold: The threshold resistance. If None, the threshold is calculated as the maximum resistance divided by 2.

    Returns:
        The critical current.
    """
    if len(current) != len(resistance):
        raise ValueError("The current and resistance arrays must have the same length.")

    if threshold is None:
        threshold = max(resistance) / 2

    for i, v in enumerate(resistance):
        if v > threshold:
            return current[i]

    return 0


def critical_temperature(temperature: Collection[Real], resistance: Collection[Real], threshold: Real | None = None) -> float:
    """
    Calculate the critical temperature from a temperature-resistance curve.

    Args:
        temperature: The temperature values.
        resistance: The resistance values.
        threshold: The threshold resistance. If None, the threshold is calculated as the maximum resistance divided by 2.

    Returns:
        The critical temperature.
    """
    if len(temperature) != len(resistance):
        raise ValueError("The temperature and resistance arrays must have the same length.")

    temperature, resistance = zip(*sorted(zip(temperature, resistance)))

    if threshold is None:
        threshold = max(resistance) / 2

    for i, r in enumerate(resistance):
        if r > threshold:
            return temperature[i]

    return 0


def ic_to_jc(ic: Real, width: Real, height: Real) -> float:
    """
    Calculate the critical current density from the critical current.

    Args:
        ic: The critical current. (mA)
        width: The width of the wire. (nm)
        height: The height of the wire. (nm)

    Returns:
        The critical current density. (kA/cm^2)
    """

    # kA/cm^2 = (10^-3 A)/(10^2 m)^2 = (10^-6 mA)/(10^-7 cm)^2

    return ic / (width * height) * 10**(-6 + 7 * 2) # kA/cm^2


def xnn_to_dataframe(xnn_path: str) -> pd.DataFrame:
    """
    Convert an XNN file to a pandas DataFrame.

    Args:
        xnn_path: The path to the XNN file.

    Returns:
        The DataFrame.
    """
    with open(xnn_path, "r") as f:
        lines = f.readlines()

    data = []

    for line in lines:
        if line.startswith("#"):
            continue

        data.append(list(map(float, line.split())))

    return pd.DataFrame(data)



from functools import partial
from qcodes import (VisaInstrument,
					validators as vals)
from qcodes.parameters import Parameter

import logging

log = logging.getLogger(__name__)


class Yokogawa7651(VisaInstrument):
	"""  
    QCoDeS driver for the Yokogawa 7651 I/V source.  

    Args:  
        VisaInstrument (_type_): _description_  

    Attributes:  
        voltage_range (Parameter):
			Set output voltage range in mV.  

        current_range (Parameter):
			Set output current range in mA.  

        voltage_limit (Parameter):
			Set output voltage limit in mV.  

        current_limit (Parameter):
			Set output current limit in mA. 

        voltage (Parameter):
			Set output voltage in mV.  

        current (Parameter):
			Set output current in mA.  

        output (Parameter):
			Output Status. ("on", "off")
    """  
	def __init__(self, name, address, **kwargs):
		# supplying the terminator means you don't need to remove it from every response
		super().__init__(name, address, terminator='\n', **kwargs)

		# init: crashes the I/O, clear from visa test panel fixes the issue
		# self.write('RC')

		
		self._status = {
			'mode': 'Voltage',
			'voltage_limit': None,
			'current_limit': None,
			'current': 0,
			'voltage': 0
			}
		''' 
			A Parameter object for the current output of the Yokogawa7651.

			Args:
				current: The current output of the Yokogawa7651. (mA)

			Returns:
				The current output of the Yokogawa7651. (mA)
		'''
		self.auto_current: Parameter = self.add_parameter(
            name="current",
            parameter_class=Y7651AutoCurrent,
            label="current",
            unit="A",
        )

		self.auto_voltage: Parameter = self.add_parameter(
			name="voltage",
			parameter_class=Y7651AutoVoltage,
			label="voltage",
			unit="V",
		)

		self.mode: Parameter = self.add_parameter(
			name="mode",
			parameter_class=Y7651Mode,
			label="mode",
		)

		self.voltage_range: Parameter = self.add_parameter(
			name = 'voltage_range',  
			label = 'Set the output voltage range in mV',
			vals = vals.Enum(10, 100, 1000, 10000, 30000),
			unit   = 'mV',
			set_cmd = partial(self._set_range, mode = "VOLT"),
			get_cmd = None
			)

		self.current_range: Parameter = self.add_parameter(
			name = 'current_range',  
			label = 'Set output current range in mA',
			vals = vals.Enum(1,10,100),
			unit   = 'A',
			set_cmd = partial(self._set_range, mode = "CURR"),
			get_cmd = None
			)

		self.voltage_limit: Parameter = self.add_parameter(
			name = 'voltage_limit',  
			label = 'Set output voltage limit in mV',
			vals = vals.Numbers(1000,30_000),
			unit   = 'mV',
			set_parser = self._div_1000_int,
			set_cmd = 'LV'+'{}'
			)

		self.current_limit: Parameter = self.add_parameter(
			name = 'current_limit',
			label = 'Set output current limit in mA',
			vals = vals.Numbers(5,120),
			unit   = 'A',
			set_parser = int,
			set_cmd = 'LA'+'{}')

		self.output = self.add_parameter(
			name = 'output',  
			label = 'Output State',
			set_cmd=lambda x: self.on() if x else self.off(),
			val_mapping={"off": 0, "on": 1,},
			)
	
	def on(self):
		self.write('O1E')
	
	def off(self):
		self.write('O0E')
	
	def _set_range(self, range:int, mode:str) -> None:
		if mode == "CURR":
			range_options = {1:"R4", 10:"R5", 100:"R6" }
			self.write('F5'+range_options[int(range)]+'E')
		elif mode == "VOLT":
			range_options = {10:"R2", 100:"R3", 1000:"R4", 10000:"R5", 30000:"R6" }
			self.write('F1'+range_options[int(range)]+'E')

	def _get_mode(self, status:str) -> str:
		if "F1R" in status:
			return "VOLT"
		elif "F5R" in status:
			return "CURR"
    
	def _get_range(self, status:str) -> int:
		if "F1R" in status:
			if "R2" in status:
				return 10
			elif "R3" in status:
				return 100
			elif "R4" in status:
				return 1000
			elif "R5" in status:
				return 10000
			elif "R6" in status:
				return 30000
		elif "F5R" in status:
			if "R4" in status:
				return 1
			elif "R5" in status:
				return 10
			elif "R6" in status:
				return 100

	def _volt_limit(self, status:str) -> int:
		if "LV" in status:
			return int(status[2:])

	def _curr_limit(self, status:str) -> int:
		if "LA" in status:
			return int(status[status.index("LA")+2:])

	def _get_status(self) -> None:
		status = self.ask('OS')
		slist = status.split()
		self.statusmap = {
			'mode': self._get_mode(slist[1]),
			'range': self._get_range(slist[1]),
			'voltage_limit': self._volt_limit(slist[3]),
			'current_limit': self._curr_limit(slist[3]),
			}

	def _div_1000_int(self,val):
		return int(val/1000)

	def _set_V(self,voltage):
		self._status['mode'] = 'Voltage'
		self._status['voltage'] = voltage
		if voltage>0:
			polarity = '+'
		else:
			polarity = '-'
		self.write('F1SA'+polarity+str(round(abs(voltage),6))+'E')

	def _set_A(self,current):
		self._status['mode'] = 'Current'
		self._status['current'] = current
		if current>0:
			polarity = '+'
		else:
			polarity = '-'
		self.write('F5SA'+polarity+str(round(abs(current),6))+'E')

	def initialize(self):
		self.write('RC')

	def reverse(self):
		self.write('RC')

	# To avoid identity query error
	def get_idn(self):
		return self.ask('OS')


class Y7651AutoCurrent(Parameter):
	def __init__(
        self,
        name: str,
        instrument: Yokogawa7651,
        **kwargs,) -> None:
		
		super().__init__(name, instrument=instrument, **kwargs)

	def get_raw(self):
		return self.instrument._status["current"]

	def set_raw(self, value):
		self.instrument._set_A(value)


class Y7651AutoVoltage(Parameter):
	def __init__(
		self,
		name: str,
		instrument: Yokogawa7651,
		**kwargs,) -> None:
		
		super().__init__(name, instrument=instrument, **kwargs)

	def get_raw(self):
		return self.instrument._status["voltage"]

	def set_raw(self, value):
		self.instrument._set_V(value)


class Y7651Mode(Parameter):
	def __init__(
		self,
		name: str,
		instrument: Yokogawa7651,
		**kwargs,) -> None:
		
		super().__init__(name, instrument=instrument, **kwargs)

	def get_raw(self):
		return self.instrument._status["mode"]

	def set_raw(self, value):
		if value == "Voltage":
			self.instrument.write('F1E')
		elif value == "Current":
			self.instrument.write('F5E')


"""
LI5650 is a QCoDeS driver for the Stanford Research Systems NF Lock-in Amplifier.

Attributes:
    phase (Parameter): Phase of the signal in degrees.
    frequency (Parameter): Frequency of the signal in Hz.
    amplitude (Parameter): Amplitude of the signal in V.
    x_offset (Parameter): X-axis offset in percentage.
    y_offset (Parameter): Y-axis offset in percentage.
    reference (Parameter): Reference signal source.
    time_constant (Parameter): Time constant of the filter in seconds.
    input_gain (Parameter): Input gain in Ohms.
    source_frequency (Parameter): Source frequency in Hz.
    filter_slope (Parameter): Filter slope in dB/octave.
    ac_sensitivity (Parameter): Current sensitivity in A.
    v_sensitivity (Parameter): Voltage sensitivity in V.
    offset_status_x (Parameter): Status of X-axis offset (ON/OFF).
    offset_status_y (Parameter): Status of Y-axis offset (ON/OFF).

Methods:
    auto_all(): Automatically adjusts all settings.
    auto_phase(): Automatically adjusts the phase.
    auto_offset(): Automatically adjusts the offset.
    auto_time_constant(): Automatically adjusts the time constant.
    auto_sensitivity(): Automatically adjusts the sensitivity.
    reset(): Resets the instrument.
    disable_front_panel(): Disables the front panel.
    enable_front_panel(): Enables the front panel.
"""
from functools import partial
import numpy as np
from typing import Any

from qcodes import VisaInstrument
from qcodes.parameters import ArrayParameter, ParamRawDataType
from qcodes.utils.validators import Numbers, Ints, Enum, Strings

from typing import Tuple

class LI5650(VisaInstrument):
    """
    This is the qcodes driver for the Stanford Research Systems NF
    Lock-in Amplifier
    """

    def __init__(self, name: str, address: str, **kwargs: Any):
        super().__init__(name, address, **kwargs)

        # Reference and phase
        self.add_parameter('phase',
                           label='Phase',
                           get_cmd='PHAS?',
                           get_parser=float,
                           set_cmd='PHAS {:.2f}',
                           unit='deg',
                           vals=Numbers(min_value=-360, max_value=729.99))

 
        self.add_parameter('frequency',
                           label='Frequency',
                           get_cmd='FREQ?',
                           get_parser=float,
                           set_cmd='FREQ {:.4f}',
                           unit='Hz',
                           vals=Numbers(min_value=1e-3, max_value=11000e3))

        self.add_parameter('amplitude',
                           label='Amplitude',
                           get_cmd=':SOUR:VOLT?',
                           get_parser=float,
                           set_cmd=':SOUR:VOLT {:.3f}',
                           unit='V',
                           vals=Numbers(min_value=0, max_value=1000000))

        self.add_parameter('x_offset',
                           label='x_offset',
                           get_cmd=':CALCulate1:OFFSet?',
                           get_parser=float,
                           set_cmd=':CALCulate1:OFFSet {:.3f}',
                           unit='%',
                           vals=Numbers(min_value=0, max_value=100))

        self.add_parameter('y_offset',
                           label='y_offset',
                           get_cmd=':CALCulate2:OFFSet?',
                           get_parser=float,
                           set_cmd=':CALCulate2:OFFSet {:.3f}',
                           unit='%',
                           vals=Numbers(min_value=0, max_value=100))

        self.add_parameter('reference',
                           label='Reference',
                           get_cmd=':ROUT2?',
                           get_parser=str,
                           set_cmd=':ROUT2 {}',
                           vals=Enum("RINP", "IOSC", "SINP"))

        self.add_parameter('time_constant',
                           label='Time constant',
                           get_parser=float,
                           get_cmd=':FILT:TCON?',
                           set_cmd=':FILT:TCON {}',
                           unit='s')


        self.add_parameter('input_gain',
                           label='Input gain',
                           get_cmd=':INP:GAIN?',
                           get_parser=str,
                           set_cmd=':INP:GAIN {}',
                           unit='Ohm',
                           val_mapping={1: "IE6\n", 100: "IE8\n"})

        self.add_parameter('source_frequency',
                           label='Source frequency',
                           get_cmd=':SOUR:FREQ?',
                           get_parser=float,
                           set_cmd=':SOUR:FREQ {:.4f}',
                           unit='Hz',
                           vals=Numbers(min_value=5e-4, max_value=2.6e5))

        self.add_parameter('filter_slope',
                           label='Filter slope',
                           get_cmd=':FILT:SLOP?',
                           get_parser=int,
                           set_cmd=':FILT:SLOP {}',
                           unit='dB/oct',
                           vals=Enum(6, 12, 18, 24))

        self.add_parameter('ac_sensitivity',
                           label='Current Sensitivity',
                           get_cmd=':CURR:AC:RANG?',
                           get_parser=float,
                           set_cmd=':CURR:AC:RANG {}',
                           unit='A',
                           vals=Numbers(min_value=10e-15, max_value=1e-6))

        self.add_parameter('v_sensitivity',
                           label='Voltage Sensitivity',
                           get_cmd=':VOLT:AC:RANG?',
                           get_parser=float,
                           set_cmd=':VOLT:AC:RANG {}',
                           unit='V',
                           vals=Numbers(min_value=1e-9, max_value=1))


        # Aux input/output
        for i in [1, 2, 3, 4]:
            self.add_parameter(f'data{i}',
                               label=f'data {i}',
                               get_cmd=partial(self._get_data, i),
                               get_parser=float,
                               unit='')

                           
        self.add_parameter(name='offset_status_x',
                           get_cmd=':CALCulate1:OFFSet:STATe?',
                           set_cmd=':CALCulate1:OFFSet:STATe {}',
                           val_mapping={'OFF': 0,
                                        'ON': 1})
        
        self.add_parameter(name='offset_status_y',
                           get_cmd=':CALCulate2:OFFSet:STATe?',
                           set_cmd=':CALCulate2:OFFSet:STATe {}',
                           val_mapping={'OFF': 0,
                                        'ON': 1})
        # Auto functions
        self.add_function('auto_all', call_cmd=':SENS:AUTO:ONCE')
        self.add_function('auto_phase', call_cmd=':PHASe:AUTO:ONCE')
        self.add_function('auto_offset', call_cmd=':CALC1:OFFS:AUTO:ONCE')
        #self.add_function('auto_offset_x', call_cmd=':CALC1:OFFS:AUTO:ONCE')
        self.add_function('auto_time_constant', call_cmd=':SENS:FILT:AUTO:ONCE')
        self.add_function('auto_sensitivity', call_cmd=':SENSE:CURR:AC:RANG:AUTO:ONCE')
        # Interface
        self.add_function('reset', call_cmd='*RST')

        self.add_function('disable_front_panel', call_cmd='OVRM 0')
        self.add_function('enable_front_panel', call_cmd='OVRM 1')


    def _get_data(self, data_number):
        if data_number==1:
            self.write(':DATA 2')
        elif data_number==2:
            self.write(':DATA 4')
        elif data_number==3:
            self.write(':DATA 8')
        elif data_number==4:
            self.write(':DATA 16')
        return self.ask('FETC?')



"""
EZGraph is a class for creating and customizing 2D graphs using Plotly.

Attributes:
    _dispname (str): The display name of the graph.
    xax_title (str): Title for the x-axis.
    yax_title (str): Title for the y-axis.
    width (int): Width of the graph.
    height (int): Height of the graph.

Methods:
    add_lines_markers(xdata, ydata, name, mode, color):
        Adds a graph with lines and markers to the figure.
    add_markers(x, y, name, color, size):
        Adds markers to the figure.
    add_line(x, y, name, color, width):
        Adds a line to the figure.
    logx():
        Sets the x-axis to logarithmic scale.
    logy():
        Sets the y-axis to logarithmic scale.
    legand_loc(x, y):
        Sets the legend location.
    title_loc(y):
        Sets the title location.
    print_methods():
        Prints all callable methods of the class.
"""
import plotly.graph_objects as go
from collections.abc import Collection

class EZGraph(go.Figure):
    '''2次元グラフを描画するためのクラス

    Args:
        dispname (str): グラフの表示名
        xax_title (str): x軸のタイトル
        yax_title (str): y軸のタイトル
        width (int): グラフの幅
        height (int): グラフの高さ
    '''

    def __init__(
        self, 
        dispname:str = "",
        xax_title:str = "",
        yax_title:str = "",
        width:int = 900, 
        height:int = 600,
    ):
        super().__init__()

        # plotly側の設定により、最初にアンダーバーを入れないとエラーを吐く
        # 修正を要検討
        self._dispname = dispname

        self.update_layout(
            title = dict(text = "<b>" + self._dispname, font = dict(size=22, color='gray'), y = 0.95),
            legend=dict(xanchor='left', yanchor='bottom', x=0.02, y=0.82),
            width=width, 
            height=height,
            xaxis=dict(title = xax_title),
            yaxis=dict(title = yax_title),
            plot_bgcolor='white')

        self.update_xaxes(
            showline=True,
            linewidth=2, mirror= True, tickfont_size = 20, title_font=dict(size=24), color='black',
            linecolor='grey',
            ticks='inside',
            ticklen=5,
            tickwidth=2,
            tickcolor='grey'
            )

        self.update_yaxes(
            showline=True,
            linewidth=2, mirror= True, tickfont_size = 20, title_font=dict(size=24), color='black',
            linecolor='grey',
            ticks='inside',
            ticklen=5,
            tickwidth=2,
            tickcolor='grey'
            )
    
    def add_lines_markers(
        self,
        xdata: Collection, 
        ydata: Collection, 
        name: str = "", 
        mode: str = "lines+markers", 
        color: str | None = None
    ):
        """Add a graph to the figure.

        Args:
            xdata (Collection):
                x-axis data

            ydata (Collection):
                y-axis data

            name (str, optional):
                name of the graph. Defaults to "".

            mode (str, optional):
                mode of the graph. Defaults to "lines+markers".

            color (str | None, optional): 
                plot color. Defaults to None (default color).
        """
        self.add_trace(
            go.Scatter(x=xdata, y=ydata, mode=mode, name=name, 
                marker=dict(size=7, color=color), line=dict(width=3.5, color=color)
                )
            )


    def add_markers(self, x: Collection, y: Collection, name: str = "", color: str | None = None, size: float = 7):
        """_summary_

        Args:
            x (Collection): _description_
            y (Collection): _description_
            name (str, optional): _description_. Defaults to "".
            color (str | None, optional): _description_. Defaults to None.
            size (int | float, optional): _description_. Defaults to 7.
        """

        self.add_trace(
            go.Scatter(x=x, y=y, mode="markers", name=name, 
                marker=dict(size=size, color=color)
                )
            )
        
    def add_line(self, x: Collection, y: Collection, name: str = "", color: str | None = None, width: float = 3.5):
        """Add a line to the figure.

        Args:
            x (Collection): 
                x-axis data
            y (Collection):
                y-axis data
            name (str, optional):
                Name of the line. Defaults to "".
            color (str | None, optional): _description_. Defaults to None.
            width (float, optional): _description_. Defaults to 3.5.
        """
        self.add_trace(
            go.Scatter(x=x, y=y, mode="lines", name=name, 
                line=dict(width=width, color=color)
                )
            )

    def logx(self):
        """x軸を対数軸にする
        """
        self.update_xaxes(type="log")
    
    def logy(self): 
        """y軸を対数軸にする
        """
        self.update_yaxes(type="log")

    def legand_loc(self, x: float = 1.02, y: float = 1):
        """凡例の位置を設定する

        Args:
            x (float, optional): . Defaults to 1.02.
            y (float, optional): _description_. Defaults to 1.
        """
        self.update_layout(legend=dict(x=x, y=y))

    def title_loc(self, y:float = 0.95):
        self.update_layout(title = dict(y=y))
    
    def print_methods(self):
        print([method for method in dir(self) if callable(getattr(self, method)) and not method.startswith("__")])



# most of the drivers only need a couple of these... moved all up here for clarity below
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import (
        Unpack,  # can be imported from typing if python >= 3.12
    )

from qcodes import validators as vals
from qcodes.instrument import (
    VisaInstrument,
    VisaInstrumentKWArgs,
)
from qcodes.parameters import Parameter
from qcodes.validators import Enum, Numbers

class Keithley2182A1ch(VisaInstrument):
    """
    Instrument Driver for Keithley2182A (1 channel, Voltage only)

    Attributes:
        nplc (Parameter): Set or get the number of power line cycles (min=0.01, max=50)
        auto_range (Parameter): Set or get the measurement range automatically (1: ON, 0: OFF)
        rel (Parameter): Enables or disables the application of
                         a relative offset value to the measurement. (1: ON, 0: OFF)
        active (Parameter): Set or get the active function. (VOLT or TEMP)
        filter (Parameter): Enables or disables the digital filter for measurements.
        amplitude (Parameter): Get the voltage (unit: V)

    """
    def __init__(
        self,
        name: str,
        address: str,
        reset: bool = False,
        **kwargs: "Unpack[VisaInstrumentKWArgs]",
    ):

        super().__init__(name, address, **kwargs)

        self._trigger_sent = False

        self.nplc: Parameter = self.add_parameter(
            "nplc",
            get_cmd="SENS:VOLT:NPLC?",
            set_cmd="SENS:VOLT:NPLC {}",
            vals=Numbers(min_value=0.01, max_value=50),
            get_parser=float
        )

        # 将来的にチャンネルを変更できるようにする
        self.auto_range: Parameter = self.add_parameter(
            "auto_range",
            get_cmd="SENS:VOLT:CHAN1:RANG:AUTO?",
            set_cmd="SENS:VOLT:CHAN1:RANG:AUTO {}"
        )

        # 将来的に「チャンネルの変更」「Bool値の入力方法」「温度かボルテージか」など変更できるようにしたい
        self.rel: Parameter = self.add_parameter(
            "rel",
            get_cmd="SENS:VOLT:REFerence:STATe?",
            set_cmd="SENS:VOLT:REFerence:STATe {}"
        )

        self.active: Parameter = self.add_parameter(
            "active",
            get_cmd=":SENS:FUNC?",
            set_cmd="SENS:FUNC {}",
            vals=Enum("VOLT", "TEMP")
        )

        self.filter: Parameter = self.add_parameter(
            "filter",
            get_cmd=":SENS:VOLT:DFILter:STAT?",
            set_cmd=":SENS:VOLT:DFILter:STAT {}"
        )

        self.amplitude: Parameter = self.add_parameter(
            "amplitude",
            get_cmd="SENS:DATA:FRES?",
            get_parser=float,
            unit="V"
        )

        self.get = self.amplitude


import pyvisa

def show_connected_visa(timeout=5000):
    rm = pyvisa.ResourceManager()
    res_list = rm.list_resources()

    for i in res_list:
        inst = rm.open_resource(i)
        inst.timeout = timeout
        try:
            print(i, ": ", inst.query("*IDN?"))
        except Exception:
            print(i, ": ", "Unknown instruments (*IDN? command failed)")

def list_visa_resources():
    return pyvisa.ResourceManager().list_resources()


"""
RealTimePlot is a class for real-time data visualization using Dash.

Attributes:
    x_label (str): Label for the x-axis.
    y_label (str): Label for the y-axis.
    x_data (list): List to store x-axis data.
    y_data (list): List to store y-axis data.
    app (dash.Dash): Dash application instance.

Methods:
    update_graph(n):
        Updates the graph with the current data.
    add(x, y):
        Adds new data points to the graph.
    run(debug):
        Starts the Dash server for real-time visualization.
"""
import dash
from dash import dcc, html
from dash.dependencies import Input, Output
import plotly.graph_objs as go

class RealTimePlot:
    def __init__(self, x_label, y_label, interval=1000):
        # 軸ラベルの初期化
        self.x_label = x_label
        self.y_label = y_label
        # データを格納するリスト
        self.x_data = []
        self.y_data = []
        
        # Dash アプリケーションの設定
        self.app = dash.Dash(__name__)
        self.app.layout = html.Div([
            dcc.Graph(id='live-graph'),
            dcc.Interval(
                id='graph-update',
                interval=interval,  # 更新間隔（ミリ秒）
                n_intervals=0
            )
        ])
        
        # グラフ更新用のコールバックを設定
        self.app.callback(
            Output('live-graph', 'figure'),
            [Input('graph-update', 'n_intervals')]
        )(self.update_graph)

    def update_graph(self, n):
        # X軸とY軸の範囲を決定
        if self.x_data:
            x_range = [min(self.x_data), max(self.x_data)]
        else:
            x_range = [0, 1]
        if self.y_data:
            y_range = [min(self.y_data), max(self.y_data)]
        else:
            y_range = [0, 1]

        # プロット用のデータ作成
        data = go.Scatter(
            x=self.x_data,
            y=self.y_data,
            mode='lines+markers'
        )
        figure = {
            'data': [data],
            'layout': go.Layout(
                xaxis={'title': self.x_label, 'range': x_range},
                yaxis={'title': self.y_label, 'range': y_range}
            )
        }
        return figure

    def add(self, x, y):
        # 受け取ったデータを追加
        self.x_data.append(x)
        self.y_data.append(y)

    def run(self, debug=True):
        # Dash サーバーを起動
        self.app.run(debug=debug)

if __name__ == '__main__':
    # サンプル実行: X軸とY軸のラベルを指定してサーバーを起動
    rtp = RealTimePlot("X 軸", "Y 軸")
    rtp.run()



import tkinter as tk
from tkinter import filedialog

from typing import Union
import time


def openpath(multichoice:bool = False) -> Union[str, tuple[str, ...]]:
    """It returns path(s) of (a) file(s) that you select on filedialog.

    Args:
        multichoice (bool, optional): Selecting multiple files. Defaults to False.

    Returns:
        Union[str, tuple[str, ...]]: file path(s)
    """
    root = tk.Tk()
    root.withdraw()
    # ファイルダイアログを開いてファイルを選択
    root.lift() # ウィンドウを最前面に表示する
    root.attributes('-topmost', True) # ウィンドウを常に最前面に表示する
    path = filedialog.askopenfilename(parent=root, multiple=multichoice)
    # ルートウィンドウを閉じる
    root.attributes('-topmost', False) # 最前面表示を解除
    root.destroy()
    return path


def opendir() -> str:
    """It returns a directory path that you select on filedialog.

    Returns:
        str: a directory path
    """
    root = tk.Tk()
    root.withdraw()
    # ディレクトリ選択ダイアログを開く
    root.lift()  # ウィンドウを最前面に表示する
    root.attributes('-topmost', True)  # ウィンドウを常に最前面に表示する
    directory = filedialog.askdirectory(parent=root)
    # ルートウィンドウを閉じる
    root.attributes('-topmost', False)  # 最前面表示を解除
    root.destroy()
    return directory


def opendir_safety() -> str:
    """It returns a directory path that you select on filedialog.

    Returns:
        str: a directory path
    """
    # tk.Tk() インスタンスの作成と管理をすべて削除します。
    # filedialog.askdirectory() が暗黙的にウィンドウを処理します。
    directory = filedialog.askdirectory()
    return directory


"""
Keithley6221 is a QCoDeS driver for the Keithley 6221 AC and DC Current Source.

Attributes:
    output (Parameter): Output state (ON/OFF).
    dc_amplitude (Parameter): DC current amplitude in A.
    dc_compliance (Parameter): DC compliance voltage in V.
    auto_range (Parameter): Auto range state (ON/OFF).
    wave_func (Parameter): Waveform function type.
    wave_amplitude (Parameter): Amplitude of the waveform in A.
    wave_frec (Parameter): Frequency of the waveform in Hz.
    wave_offset (Parameter): Offset of the waveform in A.
    wave_use_phasemarker (Parameter): Phase marker usage state (ON/OFF).
    wave_phasemarker_phase (Parameter): Phase of the phase marker in degrees.
    wave_phasemarker_line (Parameter): Line number of the phase marker.

Methods:
    waveform_arm(): Arms the current waveform function.
    waveform_start(): Starts the waveform output.
    waveform_abort(): Aborts the waveform output.
    clear(): Clears the source settings.
    on(): Turns the output ON.
    off(): Turns the output OFF.
"""
# most of the drivers only need a couple of these... moved all up here for clarity below
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from typing_extensions import (
        Unpack,  # can be imported from typing if python >= 3.12
    )

from qcodes import validators as vals
from qcodes.instrument import (
    Instrument,
    VisaInstrument,
    VisaInstrumentKWArgs,
)
from qcodes.parameters import Parameter
from qcodes.validators import Enum, Ints, MultiType, Numbers

class Keithley6221(VisaInstrument):
    """Instrument Driver for Keithley6221"""

    default_terminator = "\n"

    def __init__(
        self,
        name: str,
        address: str,
        reset: bool = False,
        **kwargs: "Unpack[VisaInstrumentKWArgs]",
    ):

        """initial
        """
        super().__init__(name, address, **kwargs)


        self.output: Parameter = self.add_parameter(
            "output",
            get_cmd="OUTP:STAT?",
            set_cmd="OUTP:STAT {}",
            vals = Enum("ON", "OFF", "0", "1")
        )

        self.dc_amplitude: Parameter = self.add_parameter(
            "dc_amplitude",
            get_cmd="SOUR:CURR:AMPL?",
            set_cmd="SOUR:CURR:AMPL {}",
            vals=Numbers(min_value=-105e-3, max_value=105e-3),
            unit="A"
        )

        self.dc_compliance: Parameter = self.add_parameter(
            "dc_compliance",
            get_cmd="SOUR:CURR:COMP?",
            set_cmd="SOUR:CURR:COMP {}",
            vals=Numbers(min_value=0.1, max_value=105),
            unit="V"
        )

        self.auto_range: Parameter = self.add_parameter(
            "auto_range",
            get_cmd="SOUR:CURR:RANG:AUTO?",
            set_cmd="SOUR:CURR:RANG:AUTO {}",
            vals = Enum("ON", "OFF", "0", "1")
        )

        self.wave_func: Parameter = self.add_parameter(
            "wave_func",
            get_cmd="SOUR:WAVE:FUNC?",
            set_cmd="SOUR:WAVE:FUNC {}",
            val_mapping = {
                "sine": "SIN",
                "ramp": "RAMP",
                "square": "SQU",
                "arbitrary1": "ARB1",
                "arbitrary2": "ARB2",
                "arbitrary3": "ARB3",
                "arbitrary4": "ARB4",
            },
        )

        self.wave_amplitude: Parameter = self.add_parameter(
            "wave_amplitude",
            get_cmd="SOUR:WAVE:AMPL?",
            set_cmd="SOUR:WAVE:AMPL {}",
            vals=Numbers(min_value=2e-12, max_value=0.105),
            unit="A"
        )

        self.wave_frec: Parameter = self.add_parameter(
            "wave_frec",
            get_cmd="SOUR:WAVE:FREQ?",
            set_cmd="SOUR:WAVE:FREQ {}",
            vals=Numbers(min_value=1e-3, max_value=1e5),
            unit="Hz"
        )

        self.wave_offset: Parameter = self.add_parameter(
            "wave_offset",
            get_cmd="SOUR:WAVE:OFFS?",
            set_cmd="SOUR:WAVE:OFFS {}",
            vals=Numbers(min_value=-105e-3, max_value=105e-3),
            unit="A"
        )

        self.wave_use_phasemarker: Parameter = self.add_parameter(
            "wave_use_phasemarker",
            get_cmd="SOUR:WAVE:PMAR:STAT?",
            set_cmd="SOUR:WAVE:PMAR:STAT {}",
            vals=Enum("0", "1")
        )

        self.wave_phasemarker_phase: Parameter = self.add_parameter(
            "wave_phasemarker_phase",
            get_cmd="SOUR:WAVE:PMAR?",
            set_cmd="SOUR:WAVE:PMAR {}",
            vals=Numbers(min_value=-180, max_value=180),
    )
        
        self.wave_phasemarker_line: Parameter = self.add_parameter(
            "wave_phasemarker_line",
            get_cmd="SOUR:WAVE:PMAR:OLIN?",
            set_cmd="SOUR:WAVE:PMAR:OLIN {}",
            vals=Enum(1, 2, 3, 4, 5, 6),
        )

    def waveform_arm(self):
        """ Arm the current waveform function. """
        self.write("SOUR:WAVE:ARM")

    def waveform_start(self):
        """ Start the waveform output. Must already be armed """
        self.write("SOUR:WAVE:INIT")

    def waveform_abort(self):
        """ Abort the waveform output and disarm the waveform function. """
        self.write("SOUR:WAVE:ABOR")

    def clear(self):
        self.write("SOUR:CLE:IMM")

    def on(self):
        self.write("OUTPUT ON")
    
    def off(self):
        self.write("OUTPUT OFF")



"""
EZGraphDisplay is a class for interactive 2D graph visualization using Jupyter widgets.

Attributes:
    df (pd.DataFrame): The input DataFrame for visualization.
    x_axis (ipywidgets.Select): Widget for selecting the x-axis column.
    y_axis (ipywidgets.Select): Widget for selecting the y-axis column.
    y_axis2 (ipywidgets.Select): Widget for selecting the secondary y-axis column.
    width (ipywidgets.IntSlider): Widget for adjusting the graph width.
    logx_checkbox (ipywidgets.Checkbox): Widget for enabling logarithmic x-axis.
    logy_checkbox (ipywidgets.Checkbox): Widget for enabling logarithmic y-axis.

Methods:
    None
"""
import pandas as pd
from ipywidgets import interact, Select, IntSlider, Checkbox

from .ezgraph_2d import EZGraph

class EZGraphDisplay:
    """DataFrameを受け取り、2次元グラフを描画するためのクラス
    """

    def _gen_graph(self, x_axis, y_axis, y_axis2, logx, logy, width):
        graph = EZGraph(xax_title=self.x_axis.value, yax_title=self.y_axis.value, width=width)
        graph.add_lines_markers(self.df[x_axis], self.df[y_axis])
        if y_axis2 != "なし":
            import plotly.graph_objects as go
            graph.add_trace(
                go.Scatter(
                    x=self.df[x_axis],
                    y=self.df[y_axis2],
                    mode="lines+markers",
                    name=str(y_axis2),
                    marker=dict(size=7),
                    line=dict(width=3.5),
                    yaxis="y2"
                )
            )
            graph.update_layout(yaxis2=dict(overlaying='y', side='right'))
        if logx:
            graph.logx()
        if logy:
            graph.logy()
        graph.show()

    def __init__(self, df: pd.DataFrame):
        self.df = df
        self.x_axis = Select(options=df.columns, description='X軸', rows=4)
        self.y_axis = Select(options=df.columns, description='Y軸', rows=4)
        self.y_axis2 = Select(options=["なし"] + list(df.columns), description='Y2軸', rows=4)
        self.width = IntSlider(value=800, min=400, max=1200, step=500, description='Width')
        self.logx_checkbox = Checkbox(value=False, description="X対数")
        self.logy_checkbox = Checkbox(value=False, description="Y対数")
        interact(
            self._gen_graph,
            x_axis=self.x_axis,
            y_axis=self.y_axis,
            y_axis2=self.y_axis2,
            logx=self.logx_checkbox,
            logy=self.logy_checkbox,
            width=self.width
        )



# pralab_phys.qcodes_drivers

from .yokogawa7651 import Yokogawa7651
from .keithley2182a1ch import Keithley2182A1ch
from .qddotnet import QDdotNET
from .keithley6221 import Keithley6221
#from .nfli5640 import nfLI5640
from .li5650 import LI5650


import clr
from qcodes.parameters import Parameter
from qcodes.instrument.base import Instrument
clr.AddReference("System")
from System import Double, UInt16, Int32


# add .net reference and import so python can see .net

DEFAULT_PORT = 11000

class QDdotNET(Instrument):
    """
    A driver class for Quantum Design instruments, providing control over temperature, magnetic field, and position.

    Parameters
    ----------
    name : str
        The name of the instrument instance.
    ip_address : str
        The IP address of the Quantum Design instrument.
    instruent_type : str, optional
        The type of the instrument (default is "DynaCool").
    remote : bool, optional
        Whether to use remote control (default is False).
    QDInstrumentdir : str, optional
        Directory path to the Quantum Design instrument DLL (default is "").
    QDInstrumentname : str, optional
        Name of the Quantum Design instrument DLL (default is "QDInstrument").
    port : int, optional
        The port number for communication (default is 11000).
    **kwargs : dict
        Additional keyword arguments passed to the base class.

    Attributes
    ----------
    temperature : QDTemperature
        Parameter for controlling and reading the temperature.
    temperaturestatus : QDTemperatureStatus
        Parameter for reading the temperature status.
    field : QDField
        Parameter for controlling and reading the magnetic field.
    fieldstatus : QDFieldStatus
        Parameter for reading the magnetic field status.
    position : QDPosition
        Parameter for controlling and reading the position.
    positionstatus : QDPositionStatus
        Parameter for reading the position status.
    fieldrate : Parameter
        Parameter for setting and getting the magnetic field rate.
    temperaturerate : Parameter
        Parameter for setting and getting the temperature rate.
    positionrate : Parameter
        Parameter for setting and getting the position rate.

    Methods
    -------
    get_field()
        Retrieves the current magnetic field value.
    set_field(field)
        Sets the magnetic field to the specified value.
    get_position()
        Retrieves the current position of the motor.
    set_position(position)
        Sets the motor position to the specified value.
    set_temperature(temp)
        Sets the temperature to the specified value.
    get_temperature()
        Retrieves the current temperature value.
    get_idn()
        Returns the instrument identification information.
    """

    def __init__(
        self,
        name: str,
        ip_address: str,
        instruent_type = "DynaCool",
        remote: bool = False,
        QDInstrumentdir: str = "",
        QDInstrumentname: str = "QDInstrument",
        port = DEFAULT_PORT,
        **kwargs,
    ) -> None:
        super().__init__(name=name, **kwargs)

        clr.AddReference(QDInstrumentdir+QDInstrumentname)
        
        from QuantumDesign.QDInstrument import QDInstrumentBase, QDInstrumentFactory

        INST_MAP = {"PPMS": QDInstrumentBase.QDInstrumentType.PPMS, "DynaCool":  QDInstrumentBase.QDInstrumentType.DynaCool}

        self.QDIBase = QDInstrumentBase
        self.QDIFactory = QDInstrumentFactory
        try:
            self.device = QDInstrumentFactory.GetQDInstrument(INST_MAP[instruent_type], remote, ip_address, UInt16(port))
        except Exception:
            raise RuntimeError("Unsupported instrument_type")
        
        self.Brate = 100
        self.Trate = 10
        self.prate = 10
        self.minimim_temperature = 1.8

        self.temperature: Parameter = self.add_parameter(
            name = "temperature",
            parameter_class=QDTemperature,
            label = "temperature",
            unit = "K"
        )

        self.temperaturestatus: Parameter = self.add_parameter(
            name = "temperaturestatus",
            parameter_class=QDTemperatureStatus,
            label = "temperaturestatus",
        )

        self.field: Parameter = self.add_parameter(
            name = "field",
            parameter_class = QDField,
            label = "field",
            unit = "T"
        )

        self.fieldstatus: Parameter = self.add_parameter(
            name = "fieldstatus",
            parameter_class = QDFieldStatus,
            label = "fieldstatus",
        )

        self.position: Parameter = self.add_parameter(
            name = "position",
            parameter_class = QDPosition,
            label = "position",
            unit = "deg"
        )

        self.positionstatus: Parameter = self.add_parameter(
            name = "positionstatus",
            parameter_class = QDPositionStatus,
            label = "positionstatus",
        )

        self.fieldrate: Parameter = self.add_parameter(
            name = "fieldrate",
            label = "fieldrate",
            unit = "Oe/s",
            get_cmd = lambda: self.Brate,
            set_cmd = self._set_b_rate,
        )

        self.temperaturerate: Parameter = self.add_parameter(
            name = "temperaturerate",
            label = "temperaturerate",
            unit = "K/s",
            get_cmd = lambda: self.Trate,
            set_cmd = self._set_t_rate,
        )

        self.positionrate: Parameter = self.add_parameter(
            name = "positionrate",
            label = "positionrate",
            unit = "deg/s",
            get_cmd = lambda: self.prate,
            set_cmd = self._set_p_rate,
        )


    def get_field(self):
        return self.device.GetField(Double(0), self.QDIBase.FieldStatus(Int32(0)))
    
    def set_field(self, field):
        if -90000 <= field <= 90000:
            return self.device.SetField(field, self.Brate, self.QDIBase.FieldApproach(Int32(0)), self.QDIBase.FieldMode(0))
        else:
            raise RuntimeError("Field is out of bounds. Should be between -90000 and 90000 Oe")

    def get_position(self):
        error, position, status = self.device.GetPosition("Horizontal Rotator", Double(0), self.QDIBase.PositionStatus(Int32(0)))
        return (error, position, status)

    def set_position(self, position):
        return self.device.SetPosition("Horizontal Rotator", Double(position), self.prate, self.QDIBase.PositionMode(Int32(0)))

    def set_temperature(self, temp):
        if self.minimim_temperature <= temp <= 350:
            return self.device.SetTemperature(temp, self.Trate, self.QDIBase.TemperatureApproach(Int32(0)))
        else:
            raise RuntimeError("Temperature is out of bounds. Should be between " + str(self.minimim_temperature) + " and 350 K")

    def get_temperature(self) -> float:
        error, temperature, status = self.device.GetTemperature(Double(0), self.QDIBase.TemperatureStatus(Int32(0)))
        return (error, temperature, status)

    def _set_t_rate(self, rate: float):
        self.Trate = rate
    
    def _set_b_rate(self, rate: float):
        self.Brate = rate

    def _set_p_rate(self, rate: float):
        self.prate = rate

    def get_idn(self):
        return {
            "vendor": "Qauntum Design",
            "model": "PPMS3",
            "serial": self.serial,
            "firmware": None,
        }
    

class QDPosition(Parameter):
    """
    Parameter class for controlling and reading the motor position.

    Parameters
    ----------
    name : str
        The name of the parameter.
    instrument : QDdotNET
        The instrument instance this parameter belongs to.
    **kwargs : dict
        Additional keyword arguments.

    Methods
    -------
    set_raw(position: float)
        Sets the motor position to the specified value.
    get_raw() -> float
        Retrieves the current motor position.
    """
    """
    Parameter class for the motor position
    """
    def __init__(
        self,
        name: str,
        instrument: QDdotNET,
        **kwargs,
    ) -> None:
        super().__init__(name, instrument=instrument, **kwargs)

    def set_raw(self, position: float) -> None:
        """Sets the motor position"""
        self.instrument.set_position(position)

    def get_raw(self) -> float:
        """Returns the motor position"""
        return self.instrument.get_position()[1]
    

class QDPositionStatus(Parameter):
    """
    Parameter class for reading the motor position status.

    Parameters
    ----------
    name : str
        The name of the parameter.
    instrument : QDdotNET
        The instrument instance this parameter belongs to.
    **kwargs : dict
        Additional keyword arguments.

    Methods
    -------
    get_raw() -> int
        Retrieves the current motor position status.
    """
    """
    Parameter class for the motor position status
    """
    def __init__(
        self,
        name: str,
        instrument: QDdotNET,
        **kwargs,
    ) -> None:
        super().__init__(name, instrument=instrument, **kwargs)

    def get_raw(self) -> float:
        """Returns the motor position status"""
        return int(self.instrument.get_position()[2])


class QDTemperature(Parameter):
    """
    Parameter class for controlling and reading the temperature.

    Parameters
    ----------
    name : str
        The name of the parameter.
    instrument : QDdotNET
        The instrument instance this parameter belongs to.
    **kwargs : dict
        Additional keyword arguments.

    Methods
    -------
    set_raw(temperature: float)
        Sets the temperature to the specified value.
    get_raw() -> float
        Retrieves the current temperature value.
    """
    """
    Parameter class for the temperature
    """
    def __init__(
        self,
        name: str,
        instrument: QDdotNET,
        **kwargs,
    ) -> None:
        super().__init__(name, instrument=instrument, **kwargs)

    def set_raw(self, temperature: float) -> None:
        """Sets the temperature"""
        self.instrument.set_temperature(temperature)

    def get_raw(self) -> float:
        """Returns the temperature"""
        return self.instrument.get_temperature()[1]


class QDTemperatureStatus(Parameter):
    """
    Parameter class for reading the temperature status.

    Parameters
    ----------
    name : str
        The name of the parameter.
    instrument : QDdotNET
        The instrument instance this parameter belongs to.
    **kwargs : dict
        Additional keyword arguments.

    Methods
    -------
    get_raw() -> int
        Retrieves the current temperature status.
    """
    """
    Parameter class for the temperature
    """
    def __init__(
        self,
        name: str,
        instrument: QDdotNET,
        **kwargs,
    ) -> None:
        super().__init__(name, instrument=instrument, **kwargs)

    def get_raw(self) -> float:
        """Returns the temperature"""
        return int(self.instrument.get_temperature()[2])


class QDField(Parameter):
    """
    Parameter class for controlling and reading the magnetic field.

    Parameters
    ----------
    name : str
        The name of the parameter.
    instrument : QDdotNET
        The instrument instance this parameter belongs to.
    **kwargs : dict
        Additional keyword arguments.

    Methods
    -------
    set_raw(mag: float)
        Sets the magnetic field to the specified value.
    get_raw() -> float
        Retrieves the current magnetic field value.
    """
    """
    Parameter class for the field
    """
    def __init__(
        self,
        name: str,
        instrument: QDdotNET,
        **kwargs,
    ) -> None:
        super().__init__(name, instrument=instrument, **kwargs)

    def set_raw(self, mag: float) -> None:
        """Sets the magnetic field"""
        self.instrument.set_field(mag)

    def get_raw(self) -> float:
        """Returns the magnetic field"""
        return self.instrument.get_field()[1]


class QDFieldStatus(Parameter):
    """
    Parameter class for the field status
    """
    def __init__(
        self,
        name: str,
        instrument: QDdotNET,
        **kwargs,
    ) -> None:
        super().__init__(name, instrument=instrument, **kwargs)

    def get_raw(self) -> float:
        """Returns the magnetic field"""
        return int(self.instrument.get_field()[2])
